@@ -45,14 +45,20 @@ public:
 
     /* search method */
     int search(pinyin_option_t options, int phrase_length,
-               /* in */ ChewingKey keys[],
+               /* in */ const ChewingKey keys[],
                /* out */ PhraseIndexRanges ranges) const;
 
     /* add/remove index method */
-    int add_index(int phrase_length, /* in */ ChewingKey keys[],
+    int add_index(int phrase_length, /* in */ const ChewingKey keys[],
                   /* in */ phrase_token_t token);
-    int remove_index(int phrase_length, /* in */ ChewingKey keys[],
+    int remove_index(int phrase_length, /* in */ const ChewingKey keys[],
                      /* in */ phrase_token_t token);
+
+    /* get length method */
+    int get_length() const;
+
+    /* mask out method */
+    bool mask_out(phrase_token_t mask, phrase_token_t value);
 };
 
 
@@ -66,7 +72,7 @@ protected:
 
     /* compress consecutive tokens */
     int convert(pinyin_option_t options,
-                ChewingKey keys[],
+                const ChewingKey keys[],
                 IndexItem * begin,
                 IndexItem * end,
                 PhraseIndexRanges ranges) const;
@@ -78,13 +84,19 @@ public:
                table_offset_t & end);
 
     /* search method */
-    int search(pinyin_option_t options, /* in */ChewingKey keys[],
+    int search(pinyin_option_t options, /* in */const ChewingKey keys[],
                /* out */ PhraseIndexRanges ranges) const;
 
     /* add/remove index method */
-    int add_index(/* in */ ChewingKey keys[], /* in */ phrase_token_t token);
-    int remove_index(/* in */ ChewingKey keys[],
+    int add_index(/* in */ const ChewingKey keys[], /* in */ phrase_token_t token);
+    int remove_index(/* in */ const ChewingKey keys[],
                      /* in */ phrase_token_t token);
+
+    /* get length method */
+    int get_length() const;
+
+    /* mask out method */
+    bool mask_out(phrase_token_t mask, phrase_token_t value);
 };
 
 };
@@ -117,14 +129,14 @@ void ChewingBitmapIndexLevel::reset() {
 /* search method */
 
 int ChewingBitmapIndexLevel::search(int phrase_length,
-                                    /* in */ ChewingKey keys[],
+                                    /* in */ const ChewingKey keys[],
                                     /* out */ PhraseIndexRanges ranges) const {
     assert(phrase_length > 0);
     return initial_level_search(phrase_length, keys, ranges);
 }
 
 int ChewingBitmapIndexLevel::initial_level_search (int phrase_length,
-    /* in */ ChewingKey keys[], /* out */ PhraseIndexRanges ranges) const {
+    /* in */ const ChewingKey keys[], /* out */ PhraseIndexRanges ranges) const {
 
 /* macros */
 #define MATCH(AMBIGUITY, ORIGIN, ANOTHER) case ORIGIN:                  \
@@ -185,7 +197,7 @@ int ChewingBitmapIndexLevel::initial_level_search (int phrase_length,
 
 
 int ChewingBitmapIndexLevel::middle_and_final_level_search
-(ChewingInitial initial, int phrase_length, /* in */ ChewingKey keys[],
+(ChewingInitial initial, int phrase_length, /* in */ const ChewingKey keys[],
  /* out */ PhraseIndexRanges ranges) const {
 
 /* macros */
@@ -256,7 +268,7 @@ int ChewingBitmapIndexLevel::middle_and_final_level_search
 
 int ChewingBitmapIndexLevel::tone_level_search
 (ChewingInitial initial, ChewingMiddle middle, ChewingFinal final,
- int phrase_length, /* in */ ChewingKey keys[],
+ int phrase_length, /* in */ const ChewingKey keys[],
  /* out */ PhraseIndexRanges ranges) const {
 
     int result = SEARCH_NONE;
@@ -336,17 +348,16 @@ ChewingLengthIndexLevel::~ChewingLengthIndexLevel() {
     }
 #undef CASE
     g_array_free(m_chewing_array_indexes, TRUE);
-    m_chewing_array_indexes = NULL;
 }
 
 
 int ChewingLengthIndexLevel::search(pinyin_option_t options, int phrase_length,
-                                    /* in */ ChewingKey keys[],
+                                    /* in */ const ChewingKey keys[],
                                     /* out */ PhraseIndexRanges ranges) const {
     int result = SEARCH_NONE;
-    if (m_chewing_array_indexes->len < phrase_length + 1)
+    if ((int) m_chewing_array_indexes->len < phrase_length + 1)
         return result;
-    if (m_chewing_array_indexes->len > phrase_length + 1)
+    if ((int) m_chewing_array_indexes->len > phrase_length + 1)
         result |= SEARCH_CONTINUED;
 
 #define CASE(len) case len:                                             \
@@ -386,7 +397,7 @@ int ChewingLengthIndexLevel::search(pinyin_option_t options, int phrase_length,
 
 template<size_t phrase_length>
 int ChewingArrayIndexLevel<phrase_length>::search
-(pinyin_option_t options, /* in */ChewingKey keys[],
+(pinyin_option_t options, /* in */ const ChewingKey keys[],
  /* out */ PhraseIndexRanges ranges) const {
     IndexItem * chunk_begin = NULL, * chunk_end = NULL;
     chunk_begin = (IndexItem *) m_chunk.begin();
@@ -412,7 +423,7 @@ int ChewingArrayIndexLevel<phrase_length>::search
 /* compress consecutive tokens */
 template<size_t phrase_length>
 int ChewingArrayIndexLevel<phrase_length>::convert
-(pinyin_option_t options, ChewingKey keys[],
+(pinyin_option_t options, const ChewingKey keys[],
  IndexItem * begin, IndexItem * end,
  PhraseIndexRanges ranges) const {
     IndexItem * iter = NULL;
@@ -460,7 +471,7 @@ int ChewingArrayIndexLevel<phrase_length>::convert
 /* add/remove index method */
 
 int ChewingBitmapIndexLevel::add_index(int phrase_length,
-                                       /* in */ ChewingKey keys[],
+                                       /* in */ const ChewingKey keys[],
                                        /* in */ phrase_token_t token) {
     const ChewingKey first_key = keys[0];
     ChewingLengthIndexLevel * & length_array = m_chewing_length_indexes
@@ -475,25 +486,34 @@ int ChewingBitmapIndexLevel::add_index(int phrase_length,
 }
 
 int ChewingBitmapIndexLevel::remove_index(int phrase_length,
-                                          /* in */ ChewingKey keys[],
+                                          /* in */ const ChewingKey keys[],
                                           /* in */ phrase_token_t token) {
     const ChewingKey first_key = keys[0];
     ChewingLengthIndexLevel * & length_array = m_chewing_length_indexes
         [first_key.m_initial][first_key.m_middle]
         [first_key.m_final][first_key.m_tone];
 
-    if (length_array)
-        return length_array->remove_index(phrase_length - 1, keys + 1, token);
-    return ERROR_REMOVE_ITEM_DONOT_EXISTS;
+    if (NULL == length_array)
+        return ERROR_REMOVE_ITEM_DONOT_EXISTS;
+
+    int retval = length_array->remove_index(phrase_length - 1, keys + 1, token);
+
+    /* remove empty array. */
+    if (0 == length_array->get_length()) {
+        delete length_array;
+        length_array = NULL;
+    }
+
+    return retval;
 }
 
 int ChewingLengthIndexLevel::add_index(int phrase_length,
-                                       /* in */ ChewingKey keys[],
+                                       /* in */ const ChewingKey keys[],
                                        /* in */ phrase_token_t token) {
     if (!(phrase_length + 1 < MAX_PHRASE_LENGTH))
         return ERROR_PHRASE_TOO_LONG;
 
-    if (m_chewing_array_indexes->len <= phrase_length)
+    if ((int) m_chewing_array_indexes->len <= phrase_length)
         g_array_set_size(m_chewing_array_indexes, phrase_length + 1);
 
 #define CASE(len) case len:                                     \
@@ -531,12 +551,12 @@ int ChewingLengthIndexLevel::add_index(int phrase_length,
 }
 
 int ChewingLengthIndexLevel::remove_index(int phrase_length,
-                                          /* in */ ChewingKey keys[],
+                                          /* in */ const ChewingKey keys[],
                                           /* in */ phrase_token_t token) {
     if (!(phrase_length + 1 < MAX_PHRASE_LENGTH))
         return ERROR_PHRASE_TOO_LONG;
 
-    if (m_chewing_array_indexes->len <= phrase_length)
+    if ((int) m_chewing_array_indexes->len <= phrase_length)
         return ERROR_REMOVE_ITEM_DONOT_EXISTS;
 
 #define CASE(len) case len:                                     \
@@ -545,8 +565,19 @@ int ChewingLengthIndexLevel::remove_index(int phrase_length,
             (m_chewing_array_indexes,                           \
              ChewingArrayIndexLevel<len> *, len);               \
         if (NULL == array)                                      \
-            return ERROR_REMOVE_ITEM_DONOT_EXISTS;                    \
-        return array->remove_index(keys, token);                \
+            return ERROR_REMOVE_ITEM_DONOT_EXISTS;              \
+        int retval = array->remove_index(keys, token);          \
+                                                                \
+        /* remove empty array. */                               \
+        if (0 == array->get_length()) {                         \
+            delete array;                                       \
+            array = NULL;                                       \
+                                                                \
+            /* shrink self array. */                            \
+            g_array_set_size(m_chewing_array_indexes,           \
+                             get_length());                     \
+        }                                                       \
+        return retval;                                          \
     }
 
     switch (phrase_length) {
@@ -575,7 +606,7 @@ int ChewingLengthIndexLevel::remove_index(int phrase_length,
 
 template<size_t phrase_length>
 int ChewingArrayIndexLevel<phrase_length>::add_index
-(/* in */ ChewingKey keys[], /* in */ phrase_token_t token) {
+(/* in */ const ChewingKey keys[], /* in */ phrase_token_t token) {
     IndexItem * begin, * end;
 
     IndexItem add_elem(keys, token);
@@ -602,7 +633,7 @@ int ChewingArrayIndexLevel<phrase_length>::add_index
 
 template<size_t phrase_length>
 int ChewingArrayIndexLevel<phrase_length>::remove_index
-(/* in */ ChewingKey keys[], /* in */ phrase_token_t token) {
+(/* in */ const ChewingKey keys[], /* in */ phrase_token_t token) {
     IndexItem * begin, * end;
 
     IndexItem remove_elem(keys, token);
@@ -637,17 +668,18 @@ bool ChewingLargeTable::load_text(FILE * infile) {
     size_t freq;
 
     while (!feof(infile)) {
-        fscanf(infile, "%s", pinyin);
-        fscanf(infile, "%s", phrase);
-        fscanf(infile, "%u", &token);
-        fscanf(infile, "%ld", &freq);
+        int num = fscanf(infile, "%256s %256s %u %ld",
+                         pinyin, phrase, &token, &freq);
+
+        if (4 != num)
+            continue;
 
         if(feof(infile))
             break;
 
         glong len = g_utf8_strlen(phrase, -1);
 
-        FullPinyinParser2 parser;
+        PinyinDirectParser2 parser;
         ChewingKeyVector keys;
         ChewingKeyRestVector key_rests;
 
@@ -890,3 +922,126 @@ store(MemoryChunk * new_chunk, table_offset_t offset, table_offset_t & end) {
     end = offset + m_chunk.size();
     return true;
 }
+
+
+/* get length method */
+
+int ChewingLengthIndexLevel::get_length() const {
+    int length = m_chewing_array_indexes->len;
+
+    /* trim trailing zero. */
+    for (int i = length - 1; i >= 0; --i) {
+        void * array = g_array_index(m_chewing_array_indexes, void *, i);
+
+        if (NULL != array)
+            break;
+
+        --length;
+    }
+
+    return length;
+}
+
+template<size_t phrase_length>
+int ChewingArrayIndexLevel<phrase_length>::get_length() const {
+    IndexItem * chunk_begin = NULL, * chunk_end = NULL;
+    chunk_begin = (IndexItem *) m_chunk.begin();
+    chunk_end = (IndexItem *) m_chunk.end();
+
+    return chunk_end - chunk_begin;
+}
+
+
+/* mask out method */
+
+bool ChewingBitmapIndexLevel::mask_out(phrase_token_t mask,
+                                       phrase_token_t value) {
+    for (int k = CHEWING_ZERO_INITIAL; k < CHEWING_NUMBER_OF_INITIALS; ++k)
+        for (int l = CHEWING_ZERO_MIDDLE; l < CHEWING_NUMBER_OF_MIDDLES; ++l)
+            for (int m = CHEWING_ZERO_FINAL; m < CHEWING_NUMBER_OF_FINALS; ++m)
+                for (int n = CHEWING_ZERO_TONE; n < CHEWING_NUMBER_OF_TONES;
+                     ++n) {
+                    ChewingLengthIndexLevel * & length_array =
+                        m_chewing_length_indexes[k][l][m][n];
+
+                    if (NULL == length_array)
+                        continue;
+
+                    length_array->mask_out(mask, value);
+
+                    if (0 == length_array->get_length()) {
+                        delete length_array;
+                        length_array = NULL;
+                    }
+                }
+    return true;
+}
+
+bool ChewingLengthIndexLevel::mask_out(phrase_token_t mask,
+                                       phrase_token_t value) {
+#define CASE(len) case len:                                     \
+    {                                                           \
+        ChewingArrayIndexLevel<len> * & array = g_array_index   \
+            (m_chewing_array_indexes,                           \
+             ChewingArrayIndexLevel<len> *, len);               \
+                                                                \
+        if (NULL == array)                                      \
+            continue;                                           \
+                                                                \
+        array->mask_out(mask, value);                           \
+                                                                \
+        if (0 == array->get_length()) {                         \
+            delete array;                                       \
+            array = NULL;                                       \
+        }                                                       \
+        break;                                                  \
+    }
+
+    for (guint i = 0; i < m_chewing_array_indexes->len; ++i) {
+        switch (i){
+	    CASE(0);
+	    CASE(1);
+	    CASE(2);
+	    CASE(3);
+	    CASE(4);
+	    CASE(5);
+	    CASE(6);
+	    CASE(7);
+	    CASE(8);
+	    CASE(9);
+	    CASE(10);
+	    CASE(11);
+	    CASE(12);
+	    CASE(13);
+	    CASE(14);
+	    CASE(15);
+	default:
+	    assert(false);
+        }
+    }
+#undef CASE
+    g_array_set_size(m_chewing_array_indexes, get_length());
+    return true;
+}
+
+template<size_t phrase_length>
+bool ChewingArrayIndexLevel<phrase_length>::mask_out
+(phrase_token_t mask, phrase_token_t value) {
+    IndexItem * begin = NULL, * end = NULL;
+    begin = (IndexItem *) m_chunk.begin();
+    end   = (IndexItem *) m_chunk.end();
+
+    for (IndexItem * cur = begin; cur != end; ++cur) {
+        if ((cur->m_token & mask) != value)
+            continue;
+
+        int offset = (cur - begin) * sizeof(IndexItem);
+        m_chunk.remove_content(offset, sizeof(IndexItem));
+
+        /* update chunk end. */
+        end = (IndexItem *) m_chunk.end();
+        --cur;
+    }
+
+    return true;
+}
