@@ -31,7 +31,7 @@
 #include "chewing_key.h"
 #include "pinyin_parser_table.h"
 #include "double_pinyin_table.h"
-#include "chewing_table.h"
+#include "zhuyin_table.h"
 
 
 using namespace pinyin;
@@ -58,56 +58,6 @@ static bool check_pinyin_options(pinyin_option_t options, const pinyin_index_ite
     return true;
 }
 
-static bool check_chewing_options(pinyin_option_t options, const chewing_index_item_t * item) {
-    guint32 flags = item->m_flags;
-    assert (flags & IS_CHEWING);
-
-    /* handle incomplete chewing. */
-    if (flags & CHEWING_INCOMPLETE) {
-        if (!(options & CHEWING_INCOMPLETE))
-            return false;
-    }
-
-    return true;
-}
-
-
-gint ChewingKey::get_table_index() {
-    assert(m_initial <  CHEWING_NUMBER_OF_INITIALS);
-    assert(m_middle < CHEWING_NUMBER_OF_MIDDLES);
-    assert(m_final < CHEWING_NUMBER_OF_FINALS);
-
-    gint index = chewing_key_table[(m_initial * CHEWING_NUMBER_OF_MIDDLES + m_middle) * CHEWING_NUMBER_OF_FINALS + m_final];
-    return index == -1 ? 0 : index;
-}
-
-gchar * ChewingKey::get_pinyin_string() {
-    assert(m_tone < CHEWING_NUMBER_OF_TONES);
-    gint index = get_table_index();
-    assert(index < G_N_ELEMENTS(content_table));
-    const content_table_item_t & item = content_table[index];
-
-    if (CHEWING_ZERO_TONE == m_tone) {
-        return g_strdup(item.m_pinyin_str);
-    } else {
-        return g_strdup_printf("%s%d", item.m_pinyin_str, m_tone);
-    }
-}
-
-gchar * ChewingKey::get_chewing_string() {
-    assert(m_tone < CHEWING_NUMBER_OF_TONES);
-    gint index = get_table_index();
-    assert(index < G_N_ELEMENTS(content_table));
-    const content_table_item_t & item = content_table[index];
-
-    if (CHEWING_ZERO_TONE == m_tone) {
-        return g_strdup(item.m_chewing_str);
-    } else {
-        return g_strdup_printf("%s%s", item.m_chewing_str,
-                               chewing_tone_table[m_tone]);
-    }
-}
-
 
 /* Pinyin Parsers */
 
@@ -168,31 +118,27 @@ static inline bool search_pinyin_index(pinyin_option_t options,
     return false;
 }
 
-static bool compare_chewing_less_than(const chewing_index_item_t & lhs,
-                                      const chewing_index_item_t & rhs){
-    return 0 > strcmp(lhs.m_chewing_input, rhs.m_chewing_input);
-}
-
-static inline bool search_chewing_index(pinyin_option_t options,
-                                        const char * chewing,
+static inline bool search_pinyin_index2(pinyin_option_t options,
+                                        const pinyin_index_item_t * index,
+                                        size_t len,
+                                        const char * pinyin,
                                         ChewingKey & key){
-    chewing_index_item_t item;
+    pinyin_index_item_t item;
     memset(&item, 0, sizeof(item));
-    item.m_chewing_input = chewing;
+    item.m_pinyin_input = pinyin;
 
-    std_lite::pair<const chewing_index_item_t *,
-                   const chewing_index_item_t *> range;
+    std_lite::pair<const pinyin_index_item_t *,
+                   const pinyin_index_item_t *> range;
     range = std_lite::equal_range
-        (chewing_index, chewing_index + G_N_ELEMENTS(chewing_index),
-         item, compare_chewing_less_than);
+        (index, index + len,
+         item, compare_pinyin_less_than);
 
     guint16 range_len = range.second - range.first;
-    assert (range_len <= 1);
-
+    assert(range_len <= 1);
     if (range_len == 1) {
-        const chewing_index_item_t * index = range.first;
+        const pinyin_index_item_t * index = range.first;
 
-        if (!check_chewing_options(options, index))
+        if (!check_pinyin_options(options, index))
             return false;
 
         key = content_table[index->m_table_index].m_chewing_key;
@@ -203,11 +149,14 @@ static inline bool search_chewing_index(pinyin_option_t options,
     return false;
 }
 
+
 /* Full Pinyin Parser */
 FullPinyinParser2::FullPinyinParser2 (){
+    m_pinyin_index = NULL; m_pinyin_index_len = 0;
     m_parse_steps = g_array_new(TRUE, FALSE, sizeof(parse_value_t));
-}
 
+    set_scheme(FULL_PINYIN_DEFAULT);
+}
 
 bool FullPinyinParser2::parse_one_key (pinyin_option_t options,
                                        ChewingKey & key,
@@ -228,13 +177,20 @@ bool FullPinyinParser2::parse_one_key (pinyin_option_t options,
             parsed_len --;
             tone_pos = parsed_len;
         }
+
+        /* check the force tone option. */
+        if (options & FORCE_TONE && CHEWING_ZERO_TONE == tone) {
+            g_free(input);
+            return false;
+        }
     }
 
     /* parse pinyin core staff here. */
 
     /* Note: optimize here? */
     input[parsed_len] = '\0';
-    if (!search_pinyin_index(options, input, key)) {
+    if (!search_pinyin_index2(options, m_pinyin_index, m_pinyin_index_len,
+                              input, key)) {
         g_free(input);
         return false;
     }
@@ -432,6 +388,26 @@ int FullPinyinParser2::final_step(size_t step_len, ChewingKeyVector & keys,
     return parsed_len;
 }
 
+bool FullPinyinParser2::set_scheme(FullPinyinScheme scheme){
+    switch(scheme){
+    case FULL_PINYIN_HANYU:
+        m_pinyin_index = pinyin_index;
+        m_pinyin_index_len = G_N_ELEMENTS(pinyin_index);
+        break;
+    case FULL_PINYIN_LUOMA:
+        m_pinyin_index = luoma_pinyin_index;
+        m_pinyin_index_len = G_N_ELEMENTS(luoma_pinyin_index);
+        break;
+    case FULL_PINYIN_SECONDARY_ZHUYIN:
+        m_pinyin_index = secondary_zhuyin_index;
+        m_pinyin_index_len = G_N_ELEMENTS(secondary_zhuyin_index);
+        break;
+    default:
+        assert(false);
+    }
+    return true;
+}
+
 bool FullPinyinParser2::post_process2(pinyin_option_t options,
                                       ChewingKeyVector & keys,
                                       ChewingKeyRestVector & key_rests,
@@ -624,6 +600,10 @@ bool DoublePinyinParser2::parse_one_key(pinyin_option_t options,
                                         const char *str, int len) const {
     options &= ~(PINYIN_CORRECT_ALL|PINYIN_AMB_ALL);
 
+    /* force tone requires at least 3 characters. */
+    if (options & FORCE_TONE && 3 != len)
+        return false;
+
     if (1 == len) {
         if (!(options & PINYIN_INCOMPLETE))
             return false;
@@ -637,15 +617,14 @@ bool DoublePinyinParser2::parse_one_key(pinyin_option_t options,
         if (NULL == sheng || strcmp(sheng, "'") == 0)
             return false;
 
-        if (search_pinyin_index(options, sheng, key)) {
+        if (search_pinyin_index(options, sheng, key))
             return true;
-        } else {
+        else
             return false;
-        }
     }
 
     ChewingTone tone = CHEWING_ZERO_TONE;
-    options &= ~(PINYIN_INCOMPLETE|CHEWING_INCOMPLETE);
+    options &= ~(PINYIN_INCOMPLETE|ZHUYIN_INCOMPLETE);
     options |= PINYIN_CORRECT_UE_VE | PINYIN_CORRECT_V_U;
 
     /* parse tone */
@@ -656,6 +635,10 @@ bool DoublePinyinParser2::parse_one_key(pinyin_option_t options,
         if (!('0' < ch && ch <= '5'))
             return false;
         tone = (ChewingTone) (ch - '0');
+
+        /* check the force tone option. */
+        if (options & FORCE_TONE && CHEWING_ZERO_TONE == tone)
+            return false;
     }
 
     if (2 == len || 3 == len) {
@@ -666,9 +649,10 @@ bool DoublePinyinParser2::parse_one_key(pinyin_option_t options,
 
         int charid = ch == ';' ? 26 : ch - 'a';
         const char * sheng = m_shengmu_table[charid].m_shengmu;
+        gchar * pinyin = NULL;
         if (NULL == sheng)
-            return false;
-        if (strcmp(sheng, "'") == 0)
+            goto fallback;
+        if (0 == strcmp(sheng, "'"))
             sheng = "";
 
         /* parse yunmu here. */
@@ -676,33 +660,60 @@ bool DoublePinyinParser2::parse_one_key(pinyin_option_t options,
         if (!IS_KEY(ch))
             return false;
 
-        charid = ch == ';' ? 26 : ch - 'a';
-        /* first yunmu */
-        const char * yun = m_yunmu_table[charid].m_yunmus[0];
-        if (NULL == yun)
-            return false;
+        do {
 
-        gchar * pinyin = g_strdup_printf("%s%s", sheng, yun);
-        if (search_pinyin_index(options, pinyin, key)) {
-            key.m_tone = tone;
+            charid = ch == ';' ? 26 : ch - 'a';
+            /* first yunmu */
+            const char * yun = m_yunmu_table[charid].m_yunmus[0];
+            if (NULL == yun)
+                break;
+
+            pinyin = g_strdup_printf("%s%s", sheng, yun);
+            if (search_pinyin_index(options, pinyin, key)) {
+                key.m_tone = tone;
+                g_free(pinyin);
+                return true;
+            }
             g_free(pinyin);
-            return true;
-        }
-        g_free(pinyin);
 
-        /* second yunmu */
-        yun = m_yunmu_table[charid].m_yunmus[1];
-        if (NULL == yun)
-            return false;
+            /* second yunmu */
+            yun = m_yunmu_table[charid].m_yunmus[1];
+            if (NULL == yun)
+                break;
 
-        pinyin = g_strdup_printf("%s%s", sheng, yun);
-        if (search_pinyin_index(options, pinyin, key)) {
-            key.m_tone = tone;
+            pinyin = g_strdup_printf("%s%s", sheng, yun);
+            if (search_pinyin_index(options, pinyin, key)) {
+                key.m_tone = tone;
+                g_free(pinyin);
+                return true;
+            }
             g_free(pinyin);
-            return true;
-        }
-        g_free(pinyin);
+        } while(0);
 
+    fallback:
+        /* support fallback table for double pinyin. */
+        if (m_fallback_table) {
+            gchar * input = g_strndup(str, 2);
+
+            const char * yunmu = NULL;
+            const double_pinyin_scheme_fallback_item_t * item =
+                m_fallback_table;
+
+            /* as the fallback table is short, just iterate the table. */
+            while(NULL != item->m_input) {
+                if (0 == strcmp(item->m_input, input))
+                    yunmu = item->m_yunmu;
+                item++;
+            }
+
+            if (NULL != yunmu && search_pinyin_index(options, yunmu, key)) {
+                key.m_tone = tone;
+                g_free(input);
+                return true;
+            }
+            g_free(input);
+            return false;
+        }
     }
 
     return false;
@@ -756,11 +767,14 @@ int DoublePinyinParser2::parse(pinyin_option_t options, ChewingKeyVector & keys,
 #undef IS_KEY
 
 bool DoublePinyinParser2::set_scheme(DoublePinyinScheme scheme) {
+    /* most double pinyin schemes doesn't use fallback table. */
+    m_fallback_table = NULL;
 
     switch (scheme) {
     case DOUBLE_PINYIN_ZRM:
-        m_shengmu_table = double_pinyin_zrm_sheng;
-        m_yunmu_table   = double_pinyin_zrm_yun;
+        m_shengmu_table  = double_pinyin_zrm_sheng;
+        m_yunmu_table    = double_pinyin_zrm_yun;
+        m_fallback_table = double_pinyin_zrm_fallback;
         return true;
     case DOUBLE_PINYIN_MS:
         m_shengmu_table = double_pinyin_mspy_sheng;
@@ -775,12 +789,14 @@ bool DoublePinyinParser2::set_scheme(DoublePinyinScheme scheme) {
         m_yunmu_table   = double_pinyin_abc_yun;
         return true;
     case DOUBLE_PINYIN_PYJJ:
-        m_shengmu_table = double_pinyin_pyjj_sheng;
-        m_yunmu_table   = double_pinyin_pyjj_yun;
+        m_shengmu_table  = double_pinyin_pyjj_sheng;
+        m_yunmu_table    = double_pinyin_pyjj_yun;
+        m_fallback_table = double_pinyin_pyjj_fallback;
         return true;
     case DOUBLE_PINYIN_XHE:
-        m_shengmu_table = double_pinyin_xhe_sheng;
-        m_yunmu_table   = double_pinyin_xhe_yun;
+        m_shengmu_table  = double_pinyin_xhe_sheng;
+        m_yunmu_table    = double_pinyin_xhe_yun;
+        m_fallback_table = double_pinyin_xhe_fallback;
         return true;
     case DOUBLE_PINYIN_CUSTOMIZED:
         assert(FALSE);
@@ -789,171 +805,101 @@ bool DoublePinyinParser2::set_scheme(DoublePinyinScheme scheme) {
     return false; /* no such scheme. */
 }
 
-/* the chewing string must be freed with g_free. */
-static bool search_chewing_symbols(const chewing_symbol_item_t * symbol_table,
-                                   const char key, const char ** chewing) {
-    *chewing = NULL;
-    /* just iterate the table, as we only have < 50 items. */
-    while (symbol_table->m_input != '\0') {
-        if (symbol_table->m_input == key) {
-            *chewing = symbol_table->m_chewing;
-            return true;
-        }
-        symbol_table ++;
-    }
-    return false;
+
+PinyinDirectParser2::PinyinDirectParser2 (){
+    m_pinyin_index = pinyin_index;
+    m_pinyin_index_len = G_N_ELEMENTS(pinyin_index);
 }
 
-static bool search_chewing_tones(const chewing_tone_item_t * tone_table,
-                                 const char key, char * tone) {
-    *tone = CHEWING_ZERO_TONE;
-    /* just iterate the table, as we only have < 10 items. */
-    while (tone_table->m_input != '\0') {
-        if (tone_table->m_input == key) {
-            *tone = tone_table->m_tone;
-            return true;
-        }
-        tone_table ++;
-    }
-    return false;
-}
+bool PinyinDirectParser2::parse_one_key(pinyin_option_t options,
+                                        ChewingKey & key,
+                                        const char *str, int len) const {
+    /* "'" are not accepted in parse_one_key. */
+    gchar * input = g_strndup(str, len);
+    assert(NULL == strchr(input, '\''));
 
+    guint16 tone = CHEWING_ZERO_TONE; guint16 tone_pos = 0;
+    guint16 parsed_len = len;
+    key = ChewingKey();
 
-bool ChewingParser2::parse_one_key(pinyin_option_t options,
-                                   ChewingKey & key,
-                                   const char *str, int len) const {
-    options &= ~(PINYIN_CORRECT_ALL|PINYIN_AMB_ALL);
-    char tone = CHEWING_ZERO_TONE;
-
-    int symbols_len = len;
-    /* probe whether the last key is tone key in str. */
     if (options & USE_TONE) {
-        char ch = str[len - 1];
-        /* remove tone from input */
-        if (search_chewing_tones(m_tone_table, ch, &tone))
-            symbols_len --;
-    }
+        /* find the tone in the last character. */
+        char chr = input[parsed_len - 1];
+        if ( '0' < chr && chr <= '5' ) {
+            tone = chr - '0';
+            parsed_len --;
+            tone_pos = parsed_len;
+        }
 
-    int i;
-    gchar * chewing = NULL; const char * onechar = NULL;
-
-    /* probe the possible chewing map in the rest of str. */
-    for (i = 0; i < symbols_len; ++i) {
-        if (!search_chewing_symbols(m_symbol_table, str[i], &onechar)) {
-            g_free(chewing);
+        /* check the force tone option. */
+        if (options & FORCE_TONE && CHEWING_ZERO_TONE == tone) {
+            g_free(input);
             return false;
         }
+    }
 
-        if (!chewing) {
-            chewing = g_strdup(onechar);
-        } else {
-            gchar * tmp = chewing;
-            chewing = g_strconcat(chewing, onechar, NULL);
-            g_free(tmp);
+    /* parse pinyin core staff here. */
+
+    /* Note: optimize here? */
+    input[parsed_len] = '\0';
+    if (!search_pinyin_index2(options, m_pinyin_index, m_pinyin_index_len,
+                              input, key)) {
+        g_free(input);
+        return false;
+    }
+
+    if (options & USE_TONE) {
+        /* post processing tone. */
+        if ( parsed_len == tone_pos ) {
+            if (tone != CHEWING_ZERO_TONE) {
+                key.m_tone = tone;
+                parsed_len ++;
+            }
         }
     }
 
-    /* search the chewing in the chewing index table. */
-    if (chewing && search_chewing_index(options, chewing, key)) {
-        /* save back tone if available. */
-        key.m_tone = tone;
-        g_free(chewing);
-        return true;
-    }
-
-    g_free(chewing);
-    return false;
+    g_free(input);
+    return parsed_len == len;
 }
 
-
-/* only characters in chewing keyboard scheme are accepted here. */
-int ChewingParser2::parse(pinyin_option_t options, ChewingKeyVector & keys,
-                          ChewingKeyRestVector & key_rests,
-                          const char *str, int len) const {
+int PinyinDirectParser2::parse(pinyin_option_t options,
+                               ChewingKeyVector & keys,
+                               ChewingKeyRestVector & key_rests,
+                               const char *str, int len) const {
     g_array_set_size(keys, 0);
     g_array_set_size(key_rests, 0);
 
-    int maximum_len = 0; int i;
-    /* probe the longest possible chewing string. */
-    for (i = 0; i < len; ++i) {
-        if (!in_chewing_scheme(options, str[i], NULL))
-            break;
-    }
-    maximum_len = i;
+    ChewingKey key; ChewingKeyRest key_rest;
 
-    /* maximum forward match for chewing. */
     int parsed_len = 0;
-    while (parsed_len < maximum_len) {
-        const char * cur_str = str + parsed_len;
-        i = std_lite::min(maximum_len - parsed_len,
-                          (int)max_chewing_length);
+    int i = 0, cur = 0, next = 0;
+    while (cur < len) {
+        /* probe next position */
+        for (i = cur; i < len; ++i) {
+            if (' ' == str[i] || '\'' == str[i])
+                break;
+        }
+        next = i;
 
-        ChewingKey key; ChewingKeyRest key_rest;
-        for (; i > 0; --i) {
-            bool success = parse_one_key(options, key, cur_str, i);
-            if (success)
+        if (parse_one_key(options, key, str + cur, next - cur)) {
+            key_rest.m_raw_begin = cur; key_rest.m_raw_end = next;
+
+            /* save the pinyin. */
+            g_array_append_val(keys, key);
+            g_array_append_val(key_rests, key_rest);
+        } else {
+            return parsed_len;
+        }
+
+        /* skip consecutive spaces. */
+        for (i = next; i < len; ++i) {
+            if (' ' != str[i] && '\'' != str[i])
                 break;
         }
 
-        if (0 == i)        /* no more possible chewings. */
-            break;
-
-        key_rest.m_raw_begin = parsed_len; key_rest.m_raw_end = parsed_len + i;
-        parsed_len += i;
-
-        /* save the pinyin. */
-        g_array_append_val(keys, key);
-        g_array_append_val(key_rests, key_rest);
+        cur = i;
+        parsed_len = i;
     }
 
     return parsed_len;
 }
-
-
-bool ChewingParser2::set_scheme(ChewingScheme scheme) {
-    switch(scheme) {
-    case CHEWING_STANDARD:
-        m_symbol_table = chewing_standard_symbols;
-        m_tone_table   = chewing_standard_tones;
-        return true;
-    case CHEWING_IBM:
-        m_symbol_table = chewing_ibm_symbols;
-        m_tone_table   = chewing_ibm_tones;
-        return true;
-    case CHEWING_GINYIEH:
-        m_symbol_table = chewing_ginyieh_symbols;
-        m_tone_table   = chewing_ginyieh_tones;
-        return true;
-    case CHEWING_ETEN:
-        m_symbol_table = chewing_eten_symbols;
-        m_tone_table   = chewing_eten_tones;
-        return true;
-    }
-
-    return false;
-}
-
-
-bool ChewingParser2::in_chewing_scheme(pinyin_option_t options,
-                                       const char key, const char ** symbol)
- const {
-    const gchar * chewing = NULL;
-    char tone = CHEWING_ZERO_TONE;
-
-    if (search_chewing_symbols(m_symbol_table, key, &chewing)) {
-        if (symbol)
-            *symbol = chewing;
-        return true;
-    }
-
-    if (!(options & USE_TONE))
-        return false;
-
-    if (search_chewing_tones(m_tone_table, key, &tone)) {
-        if (symbol)
-            *symbol = chewing_tone_table[tone];
-        return true;
-    }
-
-    return false;
-}
